@@ -19,7 +19,7 @@
 //  * pinfo->column_size, from SQLDescribeCol, does not include a NULL terminator.  For example, column_size for a
 //    char(10) column would be 10.  (Also, when dealing with SQLWCHAR, it is the number of *characters*, not bytes.)
 //
-//  * When passing a length to PyString_FromStringAndSize and similar Unicode functions, do not add the NULL
+//  * When passing a length to PyUnicode_FromStringAndSize and similar Unicode functions, do not add the NULL
 //    terminator -- it will be added automatically.  See objects/stringobject.c
 //
 //  * SQLGetData does not return the NULL terminator in the length indicator.  (Therefore, you can pass this value
@@ -66,7 +66,6 @@ inline bool IsWideType(SQLSMALLINT sqltype)
     return false;
 }
 
-// TODO: Won't pyodbc_free crash if we didn't use pyodbc_realloc.
 
 static bool ReadVarColumn(Cursor* cur, Py_ssize_t iCol, SQLSMALLINT ctype, bool& isNull, byte*& pbResult, Py_ssize_t& cbResult)
 {
@@ -78,7 +77,7 @@ static bool ReadVarColumn(Cursor* cur, Py_ssize_t iCol, SQLSMALLINT ctype, bool&
     //
     // If a non-null and non-empty value was read, pbResult will be set to a buffer containing
     // the data and cbResult will be set to the byte length.  This length does *not* include a
-    // null terminator.  In this case the data *must* be freed using pyodbc_free.
+    // null terminator.  In this case the data *must* be freed using PyMem_Free.
     //
     // If a null value was read, isNull is set to true and pbResult and cbResult will be set to
     // 0.
@@ -90,13 +89,13 @@ static bool ReadVarColumn(Cursor* cur, Py_ssize_t iCol, SQLSMALLINT ctype, bool&
     pbResult = 0;
     cbResult = 0;
 
-    const Py_ssize_t cbElement = (Py_ssize_t)(IsWideType(ctype) ? sizeof(ODBCCHAR) : 1);
+    const Py_ssize_t cbElement = (Py_ssize_t)(IsWideType(ctype) ? sizeof(uint16_t) : 1);
     const Py_ssize_t cbNullTerminator = IsBinaryType(ctype) ? 0 : cbElement;
 
     // TODO: Make the initial allocation size configurable?
     Py_ssize_t cbAllocated = 4096;
     Py_ssize_t cbUsed = 0;
-    byte* pb = (byte*)malloc((size_t)cbAllocated);
+    byte* pb = (byte*)PyMem_Malloc((size_t)cbAllocated);
     if (!pb)
     {
         PyErr_NoMemory();
@@ -180,6 +179,8 @@ static bool ReadVarColumn(Cursor* cur, Py_ssize_t iCol, SQLSMALLINT ctype, bool&
 
             cbUsed += cbRead;
 
+            TRACE("Memory Need: cbRemaining=%ld cbRead=%ld\n", (long)cbRemaining, (long)cbRead);
+
             if (cbRemaining > 0)
             {
                 // This is a tiny bit complicated by the fact that the data is null terminated,
@@ -214,7 +215,7 @@ static bool ReadVarColumn(Cursor* cur, Py_ssize_t iCol, SQLSMALLINT ctype, bool&
     }
     else
     {
-        pyodbc_free(pb);
+        PyMem_Free(pb);
     }
 
     return true;
@@ -226,10 +227,10 @@ static byte* ReallocOrFreeBuffer(byte* pb, Py_ssize_t cbNeed)
     // is freed, a memory exception is set, and 0 is returned.  Otherwise the new pointer is
     // returned.
 
-    byte* pbNew = (byte*)realloc(pb, (size_t)cbNeed);
+    byte* pbNew = (byte*)PyMem_Realloc(pb, (size_t)cbNeed);
     if (pbNew == 0)
     {
-        pyodbc_free(pb);
+        PyMem_Free(pb);
         PyErr_NoMemory();
         return 0;
     }
@@ -261,13 +262,13 @@ static PyObject* GetText(Cursor* cur, Py_ssize_t iCol)
 
     if (isNull)
     {
-        I(pbData == 0 && cbData == 0);
+        assert(pbData == 0 && cbData == 0);
         Py_RETURN_NONE;
     }
 
     PyObject* result = TextBufferToObject(enc, pbData, cbData);
 
-    pyodbc_free(pbData);
+    PyMem_Free(pbData);
 
     return result;
 }
@@ -286,26 +287,19 @@ static PyObject* GetBinary(Cursor* cur, Py_ssize_t iCol)
 
     if (isNull)
     {
-        I(pbData == 0 && cbData == 0);
+        assert(pbData == 0 && cbData == 0);
         Py_RETURN_NONE;
     }
 
     PyObject* obj;
-#if PY_MAJOR_VERSION >= 3
     obj = PyBytes_FromStringAndSize((char*)pbData, cbData);
-#else
-    obj = PyByteArray_FromStringAndSize((char*)pbData, cbData);
-#endif
-    pyodbc_free(pbData);
+    PyMem_Free(pbData);
     return obj;
 }
 
 
-static PyObject* GetDataUser(Cursor* cur, Py_ssize_t iCol, int conv)
+static PyObject* GetDataUser(Cursor* cur, Py_ssize_t iCol, PyObject* func)
 {
-    // conv
-    //   The index into the connection's user-defined conversions `conv_types`.
-
     bool isNull = false;
     byte* pbData = 0;
     Py_ssize_t cbData = 0;
@@ -314,16 +308,16 @@ static PyObject* GetDataUser(Cursor* cur, Py_ssize_t iCol, int conv)
 
     if (isNull)
     {
-        I(pbData == 0 && cbData == 0);
+        assert(pbData == 0 && cbData == 0);
         Py_RETURN_NONE;
     }
 
     PyObject* value = PyBytes_FromStringAndSize((char*)pbData, cbData);
-    pyodbc_free(pbData);
+    PyMem_Free(pbData);
     if (!value)
         return 0;
 
-    PyObject* result = PyObject_CallFunction(cur->cnxn->conv_funcs[conv], "(O)", value);
+    PyObject* result = PyObject_CallFunction(func, "(O)", value);
     Py_DECREF(value);
     if (!result)
         return 0;
@@ -332,39 +326,26 @@ static PyObject* GetDataUser(Cursor* cur, Py_ssize_t iCol, int conv)
 }
 
 
-#if PY_VERSION_HEX < 0x02060000
-static PyObject* GetDataBuffer(Cursor* cur, Py_ssize_t iCol)
-{
-    PyObject* str = GetDataString(cur, iCol);
-
-    if (str == Py_None)
-        return str;
-
-    PyObject* buffer = 0;
-
-    if (str)
-    {
-        buffer = PyBuffer_FromObject(str, 0, PyString_GET_SIZE(str));
-        Py_DECREF(str);         // If no buffer, release it.  If buffer, the buffer owns it.
-    }
-
-    return buffer;
-}
-#endif
-
 static PyObject* GetDataDecimal(Cursor* cur, Py_ssize_t iCol)
 {
-    // The SQL_NUMERIC_STRUCT support is hopeless (SQL Server ignores scale on input parameters and output columns,
-    // Oracle does something else weird, and many drivers don't support it at all), so we'll rely on the Decimal's
-    // string parsing.  Unfortunately, the Decimal author does not pay attention to the locale, so we have to modify
-    // the string ourselves.
+    // The SQL_NUMERIC_STRUCT support is hopeless (SQL Server ignores scale on input parameters
+    // and output columns, Oracle does something else weird, and many drivers don't support it
+    // at all), so we'll rely on the Decimal's string parsing.  Unfortunately, the Decimal
+    // author does not pay attention to the locale, so we have to modify the string ourselves.
     //
-    // Oracle inserts group separators (commas in US, periods in some countries), so leave room for that too.
+    // Oracle inserts group separators (commas in US, periods in some countries), so leave room
+    // for that too.
     //
-    // Some databases support a 'money' type which also inserts currency symbols.  Since we don't want to keep track of
-    // all these, we'll ignore all characters we don't recognize.  We will look for digits, negative sign (which I hope
-    // is universal), and a decimal point ('.' or ',' usually).  We'll do everything as Unicode in case currencies,
-    // etc. are too far out.
+    // Some databases support a 'money' type which also inserts currency symbols.  Since we
+    // don't want to keep track of all these, we'll ignore all characters we don't recognize.
+    // We will look for digits, negative sign (which I hope is universal), and a decimal point
+    // ('.' or ',' usually).  We'll do everything as Unicode in case currencies, etc. are too
+    // far out.
+    //
+    // This seems very inefficient.  We know the characters we are interested in are ASCII
+    // since they are -, ., and 0-9.  There /could/ be a Unicode currency symbol, but I'm going
+    // to ignore that for right now.  Therefore if we ask for the data in SQLCHAR, it should be
+    // ASCII even if the encoding is UTF-8.
 
     const TextEnc& enc = cur->cnxn->sqlwchar_enc;
     // I'm going to request the data as Unicode in case there is a weird currency symbol.  If
@@ -378,13 +359,13 @@ static PyObject* GetDataDecimal(Cursor* cur, Py_ssize_t iCol)
 
     if (isNull)
     {
-        I(pbData == 0 && cbData == 0);
+        assert(pbData == 0 && cbData == 0);
         Py_RETURN_NONE;
     }
 
     Object result(DecimalFromText(enc, pbData, cbData));
 
-    pyodbc_free(pbData);
+    PyMem_Free(pbData);
 
     return result.Detach();
 }
@@ -432,9 +413,9 @@ static PyObject* GetDataLong(Cursor* cur, Py_ssize_t iCol)
         Py_RETURN_NONE;
 
     if (pinfo->is_unsigned)
-        return PyInt_FromLong(*(SQLINTEGER*)&value);
+        return PyLong_FromLong(*(SQLINTEGER*)&value);
 
-    return PyInt_FromLong(value);
+    return PyLong_FromLong(value);
 }
 
 
@@ -520,11 +501,7 @@ static PyObject* GetUUID(Cursor* cur, Py_ssize_t iCol)
     if (cbFetched == SQL_NULL_DATA)
         Py_RETURN_NONE;
 
-#if PY_MAJOR_VERSION >= 3
     const char* szFmt = "(yyy#)";
-#else
-    const char* szFmt = "(sss#)";
-#endif
     Object args(Py_BuildValue(szFmt, NULL, NULL, &guid, (int)sizeof(guid)));
     if (!args)
         return 0;
@@ -599,18 +576,6 @@ static PyObject* GetDataTimestamp(Cursor* cur, Py_ssize_t iCol)
 }
 
 
-int GetUserConvIndex(Cursor* cur, SQLSMALLINT sql_type)
-{
-    // If this sql type has a user-defined conversion, the index into the connection's `conv_funcs` array is returned.
-    // Otherwise -1 is returned.
-
-    for (int i = 0; i < cur->cnxn->conv_count; i++)
-        if (cur->cnxn->conv_types[i] == sql_type)
-            return i;
-    return -1;
-}
-
-
 PyObject* PythonTypeFromSqlType(Cursor* cur, SQLSMALLINT type)
 {
     // Returns a type object ('int', 'str', etc.) for the given ODBC C type.  This is used to populate
@@ -624,9 +589,11 @@ PyObject* PythonTypeFromSqlType(Cursor* cur, SQLSMALLINT type)
     //
     // Keep this in sync with GetData below.
 
-    int conv_index = GetUserConvIndex(cur, type);
-    if (conv_index != -1)
-        return (PyObject*)&PyString_Type;
+    if (cur->cnxn->map_sqltype_to_converter) {
+        PyObject* func = Connection_GetConverter(cur->cnxn, type);
+        if (func)
+            return (PyObject*)&PyUnicode_Type;
+    }
 
     PyObject* pytype = 0;
     bool incref = true;
@@ -636,14 +603,7 @@ PyObject* PythonTypeFromSqlType(Cursor* cur, SQLSMALLINT type)
     case SQL_CHAR:
     case SQL_VARCHAR:
     case SQL_LONGVARCHAR:
-#if PY_MAJOR_VERSION < 3
-        if (cur->cnxn->str_enc.ctype == SQL_C_CHAR)
-            pytype = (PyObject*)&PyString_Type;
-        else
-            pytype = (PyObject*)&PyUnicode_Type;
-#else
         pytype = (PyObject*)&PyUnicode_Type;
-#endif
         break;
 
     case SQL_GUID:
@@ -654,14 +614,7 @@ PyObject* PythonTypeFromSqlType(Cursor* cur, SQLSMALLINT type)
         }
         else
         {
-#if PY_MAJOR_VERSION < 3
-            if (cur->cnxn->str_enc.ctype == SQL_C_CHAR)
-                pytype = (PyObject*)&PyString_Type;
-            else
-                pytype = (PyObject*)&PyUnicode_Type;
-#else
-            pytype = (PyObject*)&PyUnicode_Type;
-#endif
+          pytype = (PyObject*)&PyUnicode_Type;
         }
         break;
 
@@ -688,7 +641,7 @@ PyObject* PythonTypeFromSqlType(Cursor* cur, SQLSMALLINT type)
     case SQL_SMALLINT:
     case SQL_INTEGER:
     case SQL_TINYINT:
-        pytype = (PyObject*)&PyInt_Type;
+        pytype = (PyObject*)&PyLong_Type;
         break;
 
     case SQL_TYPE_DATE:
@@ -716,11 +669,7 @@ PyObject* PythonTypeFromSqlType(Cursor* cur, SQLSMALLINT type)
     case SQL_VARBINARY:
     case SQL_LONGVARBINARY:
     default:
-#if PY_VERSION_HEX >= 0x02060000
         pytype = (PyObject*)&PyByteArray_Type;
-#else
-        pytype = (PyObject*)&PyBuffer_Type;
-#endif
         break;
     }
 
@@ -739,9 +688,14 @@ PyObject* GetData(Cursor* cur, Py_ssize_t iCol)
 
     // First see if there is a user-defined conversion.
 
-    int conv_index = GetUserConvIndex(cur, pinfo->sql_type);
-    if (conv_index != -1)
-        return GetDataUser(cur, iCol, conv_index);
+    if (cur->cnxn->map_sqltype_to_converter) {
+        PyObject* func = Connection_GetConverter(cur->cnxn, pinfo->sql_type);
+        if (func) {
+            return GetDataUser(cur, iCol, func);
+        }
+        if (PyErr_Occurred())
+            return 0;
+    }
 
     switch (pinfo->sql_type)
     {
