@@ -28,13 +28,8 @@ typedef long long INT64;
 typedef unsigned long long UINT64;
 #define _strcmpi strcasecmp
 #define _strdup strdup
-#ifdef __MINGW32__
-  #include <windef.h>
-  #include <malloc.h>
-#else
   inline int max(int lhs, int rhs) { return (rhs > lhs) ? rhs : lhs; }
 #endif
-#endif
 
 #ifdef __SUN__
 #include <alloca.h>
@@ -48,6 +43,7 @@ typedef unsigned long long UINT64;
 #include <boolobject.h>
 #include <unicodeobject.h>
 #include <structmember.h>
+#include <bytesobject.h>
 
 #ifdef __CYGWIN__
 #include <windows.h>
@@ -56,16 +52,6 @@ typedef unsigned long long UINT64;
 #include <sql.h>
 #include <sqlext.h>
 
-#if PY_VERSION_HEX < 0x02050000 && !defined(PY_SSIZE_T_MIN)
-typedef int Py_ssize_t;
-#define PY_SSIZE_T_MAX INT_MAX
-#define PY_SSIZE_T_MIN INT_MIN
-#define PyInt_AsSsize_t PyInt_AsLong
-#define lenfunc inquiry
-#define ssizeargfunc intargfunc
-#define ssizeobjargproc intobjargproc
-#endif
-
 #ifndef _countof
 #define _countof(a) (sizeof(a) / sizeof(a[0]))
 #endif
@@ -122,23 +108,6 @@ inline void _strlwr(char* name)
 #define STRINGIFY(x) #x
 #define TOSTRING(x) STRINGIFY(x)
 
-// Building an actual debug version of Python is so much of a pain that it never happens.  I'm providing release-build
-// versions of assertions.
-
-#if defined(PYODBC_ASSERT) && defined(_MSC_VER)
-  #include <crtdbg.h>
-  inline void FailAssert(const char* szFile, size_t line, const char* szExpr)
-  {
-      printf("assertion failed: %s(%d)\n%s\n", szFile, (int)line, szExpr);
-      __debugbreak(); // _CrtDbgBreak();
-  }
-  #define I(expr) if (!(expr)) FailAssert(__FILE__, __LINE__, #expr);
-  #define N(expr) if (expr) FailAssert(__FILE__, __LINE__, #expr);
-#else
-  #define I(expr)
-  #define N(expr)
-#endif
-
 #ifdef PYODBC_TRACE
 void DebugTrace(const char* szFmt, ...);
 #else
@@ -146,21 +115,11 @@ inline void DebugTrace(const char* szFmt, ...) { UNUSED(szFmt); }
 #endif
 #define TRACE DebugTrace
 
-// #ifdef PYODBC_LEAK_CHECK
-// #define pyodbc_malloc(len) _pyodbc_malloc(__FILE__, __LINE__, len)
-// void* _pyodbc_malloc(const char* filename, int lineno, size_t len);
-// void pyodbc_free(void* p);
-// void pyodbc_leak_check();
-// #else
-#define pyodbc_malloc malloc
-#define pyodbc_free free
-// #endif
-
 // issue #880: entry missing from iODBC sqltypes.h
 #ifndef BYTE
   typedef unsigned char		BYTE;
 #endif
-bool pyodbc_realloc(BYTE** pp, size_t newlen);
+bool PyMem_Realloc(BYTE** pp, size_t newlen);
 // A wrapper around realloc with a safer interface.  If it is successful, *pp is updated to the
 // new pointer value.  If not successful, it is not modified.  (It is easy to forget and lose
 // the old pointer value with realloc.)
@@ -169,8 +128,4 @@ void PrintBytes(void* p, size_t len);
 const char* CTypeName(SQLSMALLINT n);
 const char* SqlTypeName(SQLSMALLINT n);
 
-#include "pyodbccompat.h"
-
-#define HERE printf("%s(%d)\n", __FILE__, __LINE__)
-
 #endif // pyodbc_h
